/*
 *	Copyright (C) 2007. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvapplication.cpp
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#include <QDebug>
#include <QMetaType>
#include <QSet>

#include <qvcore/qvapplication.h>
#include <qvgui/qvisioninterface.h>
#include <qvcore/qvpropertyholder.h>

QVApplication::QVApplication (int &argc,char **argv, QString infoString,bool GUIenabled) : QApplication(argc,argv,GUIenabled), info(infoString), unusedArguments(), qvps(), visionInterface(NULL), isRunningFlag(FALSE)
	{
	qRegisterMetaType< QVariant >("QVariant");
	qRegisterMetaType< QVCamera::TCameraStatus >("QVCamera::TCameraStatus");
	qRegisterMetaType< QVWorker::TWorkerStatus >("QVWorker::TWorkerStatus");
	qRegisterMetaType< QVImage<uChar,1> >("QVImage<uChar,1>");
	qRegisterMetaType< QVImage<sShort,1> >("QVImage<sShort,1>");
	qRegisterMetaType< QVImage<sFloat,1> >("QVImage<sFloat,1>");
	qRegisterMetaType< QVImage<uChar,3> >("QVImage<uChar,3>");
	qRegisterMetaType< QVImage<sShort,3> >("QVImage<sShort,3>");
	qRegisterMetaType< QVImage<sFloat,3> >("QVImage<sFloat,3>");

	unusedArguments = arguments();
	unusedArguments.removeAt(0); // Application name removed.
	}

int QVApplication::exec()
	{
	qDebug() << "QVApplication::exec()";

	// If --help parameter was given, show help and exit:
	if(unusedArguments.contains("--help")) 
		{
		printHelp();
		return 0;
		}

	// An initialization error of any QVPropertyHolder aborts execution:
	QSetIterator<QVPropertyHolder *> iq(qvps);
	while (iq.hasNext())
		{
		QVPropertyHolder* qvp = iq.next();
		QString lastError;
		if((lastError = qvp->getLastError()) != QString())
			{
			std::cerr << "Error initializing QVApplication: "
					  << qPrintable(lastError) << std::endl;
			return -1;
			}
		}

	// If there are unused arguments, show error and exit
	if(not unusedArguments.isEmpty())
		{
		QListIterator<QString> i(unusedArguments);
		while (i.hasNext())
			std::cerr << "Error initializing QVApplication: "
					  << "unknown command line parameter: "
					  << qPrintable(i.next()) << std::endl;
		return -1;
		}

	// Now we will open all cameras:
	iq.toFront();
	while (iq.hasNext())
		{
		QVPropertyHolder* qvp = iq.next();
		QVCamera* camera;
		if((camera = dynamic_cast<QVCamera*>(qvp)) != NULL)
			if(camera->isClosed())
				if(not camera->openCam())
					{
					std::cerr << "Error initializing QVApplication: "
							  << "could not open camera: "
							  << qPrintable(camera->getName()) << std::endl;
					return -1;
					}
		}
	qDebug() << "QVApplication::exec(): cameras opened";

	// Connect and send an initial signal for the QVThreads to start running
	// later, once the GUI is up and runnig (the initWorkers() slot will be
	// lately called by the exec() method):
	connect(this,SIGNAL(inited()),this,SLOT(initWorkers()));
	emit inited();

	qDebug() << "Entering in QApplication::exec()";
	isRunningFlag = TRUE;
	int returnvalue = QApplication::exec();
	qDebug() << "Back from QApplication::exec()";

	qDebug() << "QVApplication::exec() <- return";
	return returnvalue;
}

QStringList QVApplication::getUnusedArguments()
	{ return unusedArguments; }

void QVApplication::setArgumentAsUsed(QString argument)
	{
	qDebug() << "QVApplication::setArgumentAsUsed(QString,bool)";
	int index = unusedArguments.indexOf(argument);
	if(index != -1)
		unusedArguments.removeAt(index);
	qDebug() << "QVApplication::setArgumentAsUsed(QString,bool) <- return";
	}

void QVApplication::registerQVPropertyHolder(QVPropertyHolder *qvp)
	{
	qDebug() << "QVApplication::registerQVPropertyHolder(QVWorker)";
	qvps.insert(qvp);
	qDebug() << "QVApplication::registerQVPropertyHolder(QVWorker) -> return";
	}

void QVApplication::deregisterQVPropertyHolder(QVPropertyHolder *qvp)
	{
	qDebug() << "QVApplication::registerQVPropertyHolder(QVWorker)";
	qvps.remove(qvp);
	qDebug() << "QVApplication::registerQVPropertyHolder(QVWorker) -> return";
	}

void QVApplication::registerGUI(QVisionInterface *visionInterface) 
	{
	this->visionInterface = visionInterface;
	}

void QVApplication::printHelp() const
	{
	qDebug() << "QVApplication::printHelp()";

	std::cout << "Usage: " << qPrintable(arguments().first())
			  << " [OPTIONS]" << std::endl;
	if (info != QString())
		std::cout << qPrintable(info) << std::endl;
	std::cout << std::endl;
	QSetIterator<QVPropertyHolder *> iq(qvps);
	while (iq.hasNext())
		{
		QString infoHolder = iq.next()->infoInputProperties();
		if(infoHolder != QString() )
			std::cout << qPrintable(infoHolder) << std::endl;
		}
	qDebug() << "QVApplication::printHelp() <~ return";
	}

void QVApplication::quit()
	{
	qDebug() << "QVApplication::quit()";
	// We order all workers to finish...
	QSetIterator<QVPropertyHolder *> iq(qvps);
	while (iq.hasNext())
		{
		QVPropertyHolder* qvp = iq.next();
		QVWorker* worker;
		if((worker = dynamic_cast<QVWorker*>(qvp)) != NULL)
			worker->finish();
		}
	iq.toFront();
	// ... and then wait for all of them (Warning, it won't work if we try to
	// finish and wait in the same loop).
	while (iq.hasNext())
		{
		QVPropertyHolder* qvp = iq.next();
		QVWorker* worker;
		if((worker = dynamic_cast<QVWorker*>(qvp)) != NULL)
			// Needed to treat possible pending Qt::BlockingQueuedConnection
			// signals from qvpropertyholder.h:
			while(not worker->wait(10/*ms*/)) processEvents();
		}
	qDebug() << "QVApplication::quit(): workers finished";

	// Now we will close all cameras:
	iq.toFront();
	while (iq.hasNext())
		{
		QVPropertyHolder* qvp = iq.next();
		QVCamera* camera;
		if((camera = dynamic_cast<QVCamera*>(qvp)) != NULL)
			if(!camera->isClosed())
				camera->closeCam();
		}
	qDebug() << "QVApplication::finish(): cameras closed";

	this->exit(0);
	}


void QVApplication::initWorkers()
{
	qDebug() << "QVApplication::initWorkers()";
	QSetIterator<QVPropertyHolder *> iq(qvps);
	while (iq.hasNext())
		{
		QVPropertyHolder* qvp = iq.next();
		QVWorker* worker;
		if((worker = dynamic_cast<QVWorker*>(qvp)) != NULL)
			{
				worker->start();
			}
		}
	qDebug() << "QVApplication::initWorkers() <~ return";
}



/* Antiguo QVApplication::ready:

		qDebug() << "QVApplication::ready(): opening cameras";
		QSetIterator<QVCamera *> ic(cameras);
		while (ic.hasNext())
			{
			// TODO: AQUI ABRIMOS LAS CÁMARAS
			QVCamera * camera = ic.next();

			if (camera->openCam() == -1)
				{
				returnValue = false;
				std::cerr << "Error: could not open camera" << qPrintable(camera->getName()) << std::endl;
				}
			else
				connect(camera,SIGNAL(camClosed()),this,SLOT(finish()));
			}
		}


			// Para cada propiedad que contenga una qvimage, la publicamos en el interfaz
  			QStringListIterator uCharC1Iterator(worker->getPropertiesByType< QVImage<uChar,1> >());
				while (uCharC1Iterator.hasNext())
					{
					QString property = uCharC1Iterator.next();
					if (worker->isVisibleProperty(property))
						visionInterface->publishQVImage(worker, property);
					}

  			QStringListIterator uCharC3Iterator(worker->getPropertiesByType< QVImage<uChar,3> >());
				while (uCharC3Iterator.hasNext())
					{
					QString property = uCharC3Iterator.next();
					if (worker->isVisibleProperty(property))
						visionInterface->publishQVImage(worker, property);
					}
			}
*/
