/*
 *	Copyright (C) 2007. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvimagebuffer.h
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#ifndef QVIMAGE_BUFFER_H
#define QVIMAGE_BUFFER_H

#include <stdio.h>
#include <string.h>
#include <math.h>

#include <QObject>
#include <QRect>
#include <QDebug>
#include <ipp.h>

#include <QPoint>
#include <QSharedData>

/// \cond
typedef unsigned char 	uChar;		// Ipp8u	1 bytes
typedef unsigned short 	uShort;		// Ipp16u	2 "
typedef unsigned int 	uInt;		// Ipp32u	4 "
typedef signed char 	sChar;		// Ipp8s	1 ...
typedef signed short 	sShort;		// Ipp16s	2
typedef signed int 	sInt;		// Ipp32s	4
//typedef __INT64 	sLong;		// Ipp64s	8 (not estandar type)
//typedef __UINT64 	uLong;		// Ipp64u	8 (not estandar type)
typedef float 		sFloat;		// Ipp32f	4
typedef double 		sDouble;	// Ipp64f	8

template <typename Type = uChar, int Planes = 1> class QVImageBuffer: public QSharedData
	{
	public:
		//QVImageBuffer();

		QVImageBuffer(): QSharedData() {};
		QVImageBuffer(const QVImageBuffer<Type, Planes> &);
		QVImageBuffer(uInt cols, uInt rows, uInt step = 0, const Type * buffer = NULL);

		~QVImageBuffer()	{ delete this->data; }

		uInt getRows()		const	{ return rows; }
		uInt getCols()		const	{ return cols; }
		uInt getStep()		const	{ return step; }
		uInt getPlanes()	const	{ return planes; }
		uInt getTypeSize()	const	{ return typeSize; }

		int getDataSize()			const	{ return dataSize; }
		const Type *	const getReadData()	const	{ return data; }
		Type * const	getWriteData()		const	{ return data; }
		/*QVImageBuffer * subscribe()			{ references++; return this; }
		const uInt unsubscribe()
			{
			Q_ASSERT_X(references > 0,"QVImageBuffer::unsubscribe()","no QVImage is susbscribed");
			return --references;
			}*/

		// DEPRECATED
		//void copy(const Type * buffer, uInt step);
		//bool soleOwner()			const	{ return references < 2; }

		// WARNING: use this function with care.
		//void incrementReferences()			{ references++; }

		//QVImageBuffer<Type,Planes> & operator=(const QVImageBuffer<uChar, 1> &sourceImageBuffer);
		//QVImageBuffer<Type,Planes> & operator=(const QVImageBuffer<uChar, 3> &sourceImageBuffer);
		//QVImageBuffer<Type,Planes> & operator=(const QVImageBuffer<sShort, 1> &sourceImageBuffer);
		//QVImageBuffer<Type,Planes> & operator=(const QVImageBuffer<sShort, 3> &sourceImageBuffer);

	protected:
		Type * data;
		uInt cols, rows, step, planes, dataSize, typeSize, references;
		void allocData(uInt cols, uInt rows, uInt step = 0);
	};

template <typename Type, int Planes>
void QVImageBuffer<Type,Planes>::allocData(uInt cols, uInt rows, uInt step)
	{
	qDebug() << "QVImageBuffer<Type,Planes>::allocData(" << cols << "," << rows << "," << step << ")";
	Q_ASSERT_X(cols > 0, "QVImageBuffer::allocData()", "cols <= 0");
	Q_ASSERT_X(rows > 0, "QVImageBuffer::allocData()", "rows <= 0");
	Q_ASSERT_X((step == 0) || (step >= cols), "QVImageBuffer::allocData()", "0 < step < cols");
	this->rows = rows;
	this->cols = cols;
	this->references = 0;
	this->planes = Planes;
	this->typeSize = sizeof(Type);

	qDebug() << "QVImageBuffer<Type,Planes>::allocData(): planes" << this->planes;
	uInt temp = this->planes * this->typeSize * cols;

	if (step == 0)
		this->step = 8*(uInt)ceil((double)(temp)/8);
	else
		this->step = step;

  	Q_ASSERT_X( (this->step % 8) == 0, "QVImageBuffer::QVImageBuffer()","step % 8 != 0");
  	Q_ASSERT_X( this->step >= temp, "QVImageBuffer::QVImageBuffer()","step insufficient");

	qDebug() << "QVImageBuffer<Type,Planes>::allocData(): rows" << this->rows;

	qDebug() << "QVImageBuffer<Type,Planes>::allocData(): planes" << this->planes;
	this->dataSize = this->rows * this->step * this->planes;

	qDebug() << "QVImageBuffer<Type,Planes>::allocData(): dataSize" << this->dataSize;

	if ( (this->data != NULL) && (this->cols * this->rows != cols * rows) )
		{
		delete this->data;
		this->data = NULL;
		}

	//qDebug() << "QVImageBuffer<Type,Planes>::allocData(): dataSize" << this->dataSize;

	if (this->data == NULL)
		this->data = new Type[this->dataSize];

	qDebug() << "QVImageBuffer<Type,Planes>::allocData(): step" << this->step;
	qDebug() << "QVImageBuffer<Type,Planes>::allocData(): step" << this->getStep();
	qDebug() << "QVImageBuffer<Type,Planes>::allocData(): dataSize" << this->dataSize;
	qDebug() << "QVImageBuffer<Type,Planes>::allocData(): dataSize" << this->getDataSize();
	qDebug() << "QVImageBuffer<Type,Planes>::allocData(" << cols << "," << rows << "," << step << ") <- return";
	}

template <typename Type, int Planes>
QVImageBuffer<Type,Planes>::QVImageBuffer(uInt cols, uInt rows, uInt step, const Type *buffer): QSharedData(),
	data(NULL)
	{
	qDebug() << "QVImageBuffer<Type, Planes>::QVImageBuffer(" << cols << "," << rows << "," << step << ")";
	Q_ASSERT_X(cols > 0, "QVImageBuffer::allocData()", "cols <= 0");
	Q_ASSERT_X(rows > 0, "QVImageBuffer::allocData()", "rows <= 0");
	Q_ASSERT_X((step == 0) || (step >= cols), "QVImageBuffer::allocData()", "0 < step < cols");

	allocData(cols, rows, step);

	if (buffer != NULL)
		// *_* to *_* (copy)
		memcpy(this->data,buffer,this->dataSize);
	qDebug() << "QVImageBuffer<Type, Planes>::QVImageBuffer() <- return";
	}
/// \endcond
#endif
