/*
 *	Copyright (C) 2007. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvcomponenttree.cpp
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#include <qvdta/qvcomponenttree.h>
#define	MIN(X,Y)	(((X)>(Y))?(Y):(X))
#define	MAX(X,Y)	(((X)>(Y))?(X):(Y))

namespace qvdta
{
uInt debug_aux_print_componentTree(const QVComponentTree *componentTree, int node, int numTabs)
	{
	Q_ASSERT_X(numTabs < 8, "debug_aux_print_componentTree", "Can't show trees deeper than 8 level.");
	char tabs[16] = "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t";
	tabs[numTabs] = '\0';

	uInt	minThreshold = componentTree->nodes[node].firstThreshold,
		maxThreshold = componentTree->nodes[node].lastThreshold;

	qDebug() << "debug_print_componentTree():" << tabs << "* Node" << node;
	qDebug() << "debug_print_componentTree():" << tabs << "  Seed at ("
		<< componentTree->nodes[node].seedX << componentTree->nodes[node].seedY << ")";
	qDebug() << "debug_print_componentTree():" << tabs << "  Areas min/max ["
		<< componentTree->nodes[node].area[minThreshold] << ","
		<< componentTree->nodes[node].area[maxThreshold] << "]";
	qDebug() << "debug_print_componentTree():" << tabs << "  Thresholds min/max ["
		<< minThreshold << "," << maxThreshold << "]"; 
	qDebug() << "debug_print_componentTree():" << tabs << "  # childs [" << componentTree->nodes[node].numChilds << "]"; 

	uInt sum = 1;
	for (	uInt child = componentTree->nodes[node].child, numChilds = 1;
		child != NULL_NODE;
		child = componentTree->nodes[child].brother, numChilds++ )
		{
		Q_ASSERT(child < componentTree->numNodes);
		Q_ASSERT(numChilds <= componentTree->nodes[node].numChilds);
		sum += debug_aux_print_componentTree(componentTree, child, numTabs+1);
		}

	return sum;
	}

uInt debug_print_componentTree(const QVComponentTree *componentTree)
	{
	qDebug() << "debug_print_componentTree()";
	qDebug() << "debug_print_componentTree(): numNodes =" << componentTree->numNodes;
	qDebug() << "debug_print_componentTree(): freePoints =" << componentTree->freePoints;
	qDebug() << "debug_print_componentTree(): totalPoints =" << componentTree->totalPoints;
	qDebug() << "debug_print_componentTree(): leafNodes =" << componentTree->leafNodes;
	qDebug() << "debug_print_componentTree(): rootNode =" << componentTree->rootNode;

	uInt numPrintedNodes = debug_aux_print_componentTree(componentTree, componentTree->rootNode, 1);

	qDebug() << "debug_print_componentTree(): numPrintedNodes =" << numPrintedNodes;
	qDebug() << "debug_print_componentTree() <~ return";
	return numPrintedNodes;
	}

#define	NODE_ID(PointID)	( this->nodeID[PointID] )
#define	NODE(NodeID)		( this->nodes[NodeID] )

#define	NEW_NODE(newNodeID, SeedX, SeedY, Threshold)		\
	{							\
	newNodeID = this->numNodes++;			\
	NODE_ID(disjointSet.index(SeedX, SeedY)) = newNodeID;	\
								\
	NODE(newNodeID).seedX = SeedX;				\
	NODE(newNodeID).seedY = SeedY;				\
	NODE(newNodeID).firstThreshold =			\
	NODE(newNodeID).lastThreshold = Threshold;		\
	NODE(newNodeID).child =					\
	NODE(newNodeID).brother = NULL_NODE;			\
	NODE(newNodeID).numChilds = 0;				\
	NODE(newNodeID).area[Threshold] = 0;			\
	NODE(newNodeID).inited = false;				\
	}

#define SWAP(Type, Value1, Value2)	\
	{				\
	Type temp = Value1;		\
	Value1 = Value2;		\
	Value2 = temp;			\
	}

#define ADD_CHILD(ParentNodeID, ChildNodeID)			\
	{							\
	NODE(ChildNodeID).brother = NODE(ParentNodeID).child;	\
	NODE(ParentNodeID).child = ChildNodeID;			\
	NODE(ParentNodeID).numChilds++;				\
	}

QVComponentTree::QVComponentTree(uInt cols, uInt rows, const QVImage<uChar,1> image): numNodes(0), freePoints(0)
	{
	maxNodes = cols*rows;
	nodes = new QVComponentTreeNode[maxNodes];
	nodeID = new uInt[maxNodes];
	valid = new bool[maxNodes];
	for(uInt i=0; i<maxNodes; i++)
		{
		nodeID[i] = NULL_NODE;
		valid[i] = false;
		}

	getComponentTree(image);
	}

QVComponentTree::~QVComponentTree()
	{
	delete nodes;
	delete nodeID;
	delete valid;
	}

void QVComponentTree::pruneLowComponentTreeAux(QVImage<uChar> &image, uInt minArea, uInt node, uInt validThreshold)
	{
	int	firstThreshold = NODE(node).firstThreshold,
		lastThreshold = NODE(node).lastThreshold;

	Q_ASSERT(NODE(node).area[firstThreshold] != 0);
	Q_ASSERT(NODE(node).area[lastThreshold] != 0);
	Q_ASSERT(validThreshold >= lastThreshold);

	bool prune = false;
	int lastValidThreshold = validThreshold;

	// Aqui decidimos si podamos directamente el nodo
	// O bien si hay algún nivel al que deberíamos podar
	for (int threshold = lastThreshold; threshold >= firstThreshold && !prune; threshold--)
		if (NODE(node).area[threshold] > 0)
			{
			if (NODE(node).area[threshold] < minArea)
				prune = true;
			else
				lastValidThreshold = threshold;
			}

	// Podamos, o seguimos con los hijos
	if (prune)
		myFloodFill(image, NODE(node).seedX, NODE(node).seedY, lastValidThreshold, 0, lastValidThreshold-1);
	else
		for (uInt child = NODE(node).child; child != NULL_NODE; child = NODE(child).brother)
			pruneLowComponentTreeAux(image, minArea, child, lastValidThreshold);
	}

void QVComponentTree::pruneLowRegions(QVImage<uChar> &image, uInt minArea)
	{
	qDebug() << "pruneComponentTree()";

	uInt lastThreshold = NODE(this->rootNode).lastThreshold;

	if(NODE(this->rootNode).area[lastThreshold] <= minArea)
		return;

	pruneLowComponentTreeAux(image, minArea, this->rootNode, lastThreshold);
	qDebug() << "pruneComponentTree() <~ return";
	}

/////////////////////////////////////////////////////////////////////////////////
void QVComponentTree::pruneHighComponentTreeAux(QVImage<uChar> &image, uInt minArea, uInt node, uInt validThreshold)
	{
	int	firstThreshold = NODE(node).firstThreshold,
		lastThreshold = NODE(node).lastThreshold;

	Q_ASSERT(NODE(node).area[firstThreshold] != 0);
	Q_ASSERT(NODE(node).area[lastThreshold] != 0);
	Q_ASSERT(validThreshold >= lastThreshold);

	bool prune = false;
	int lastValidThreshold = validThreshold;

	// Aqui decidimos si podamos directamente el nodo
	// O bien si hay algún nivel al que deberíamos podar
	for (int threshold = lastThreshold; threshold >= firstThreshold && !prune; threshold--)
		if (NODE(node).area[threshold] > 0)
			{
			if (NODE(node).area[threshold] < minArea)
				prune = true;
			else
				lastValidThreshold = threshold;
			}

	// Podamos, o seguimos con los hijos
	if (prune)
		myFloodFill(image, NODE(node).seedX, NODE(node).seedY, 255-lastValidThreshold, 255-lastValidThreshold+1, 255-0);
	else
		for (uInt child = NODE(node).child; child != NULL_NODE; child = NODE(child).brother)
			pruneHighComponentTreeAux(image, minArea, child, lastValidThreshold);
	}

void QVComponentTree::pruneHighRegions(QVImage<uChar> &image, uInt minArea)
	{
	qDebug() << "pruneComponentTree()";

	uInt lastThreshold = NODE(this->rootNode).lastThreshold;

	if(NODE(this->rootNode).area[lastThreshold] <= minArea)
		return;

	pruneHighComponentTreeAux(image, minArea, this->rootNode, lastThreshold);
	qDebug() << "pruneComponentTree() <~ return";
	}

/////////////////////////////////////////////////////////////////////////////////
/// Procesa una imagen y obtiene un árbol de componentes para ella, en una dirección, desde los píxeles de
/// menos valor de gris, hasta los píxeles de más valor de gris.
/// Procesa una imagen y obtiene un árbol de componentes para ella, en una dirección, desde los píxeles de menos
/// valor de gris, hasta los píxeles de más valor de gris.
void QVComponentTree::getComponentTree(const QVImage<uChar> &image)
	{
	qDebug() << "getComponentTree()";
	uInt cols = image.getCols(), rows = image.getRows();

	QVector< QVector< QPoint > > points = qvdta::CountingSort(image);
	qvdta::QVDisjointSet disjointSet(cols, rows);

	this->numNodes = 0;
	this->leafNodes = 0;
	this->freePoints = 0;
	this->totalPoints = 0;

	for (uInt i=0; i< cols*rows; i++)
		Q_ASSERT(NODE_ID(i) == NULL_NODE);

	// Este bucle organiza la estructura del arbol de componentes, apoyandose en el disjoint set, y recorriendo los píxeles
	// en orden según su valor de gris gracias al conjunto de puntos 'points', obtenido mediante la función 'CountingSort'.
	QVIMAGE_INIT_READ(uChar,image);
	for (int threshold = 0; threshold < points.size(); threshold++)
		{
		// Actualizamos conjuntos sobre el disjoint set.
		// Unimos en el disjoint set los pixels cuyo valor de gris es igual al threshold, con las regiones, o
		// puntos colindantes, cuyo valor sea igual o menor que el threshold. Esto se hace a priori sólo sobre el disjoint
		// set.
		//
		// También aqui se unen los nodos del arbol de componentes tales que tengan en común uno de los puntos procesados.
		//
		for (int n=0; n< points[threshold].size(); n++)
			{
			QPoint p = points[threshold][n];
			uInt col = p.x(), row = p.y();
			uChar actualPixel = QVIMAGE_PIXEL(image, col, row,0);

			// recorremos la vecindad en busca de puntos cercanos para unir con el actual...
			for (uInt i = (uInt) MAX(0,(int)col-1); i< MIN(cols, col+2); i++)
				for (uInt j = (uInt) MAX(0,(int)row-1); j< MIN(rows, row+2); j++)
					if ((col != i) || (row != j))
					{
					//uInt	vecinoID = disjointSet.index(i, j);
					uInt	vecinoSet = disjointSet.find(i,j), actualSet = disjointSet.find(col, row);
					uChar	vecinoPixel = QVIMAGE_PIXEL(image, i, j,0);

					if ( (vecinoPixel <= actualPixel) && (vecinoSet != actualSet) )
						{
						// Tenemos un pixel al que debemos añadir al grupo del vecino
						uInt	newPointID = disjointSet.unify(col, row, i, j);
							
						Q_ASSERT(disjointSet.find(disjointSet.index(col, row)) == disjointSet.find(disjointSet.index(i, j)));

						Q_ASSERT( (actualSet == newPointID) || (vecinoSet == newPointID) );

						uInt	actualNodeID = NODE_ID(actualSet),
							vecinoNodeID = NODE_ID(vecinoSet);

						// Si el nuevo es NULL_POINTER, le asignamos el punto q no es NULL_POINTER 
						if (actualNodeID == NULL_NODE)
							NODE_ID(newPointID) = vecinoNodeID;
						else if (vecinoNodeID == NULL_NODE)
							NODE_ID(newPointID) = actualNodeID;
						// En caso contrario, tenemos que actual y vecino tienen nodos asociados.
						// Procedemos a crear un nuevo nodo, y a añadir los otros dos nodos a dicho
						// nuevo nodo.
						else	{
							Q_ASSERT(actualNodeID != NULL_NODE);
							Q_ASSERT(vecinoNodeID != NULL_NODE);

							// Hay que comprobar que entre los dos nodos, no haya uno recién creado
							// (area[threshold] == 0)
							// en cuyo caso, ese será el nodo padre, al que añadir el otro.
							if (!NODE(actualNodeID).inited && NODE(vecinoNodeID).inited)
								// Directamente añadimos el otro nodo 
								{
								ADD_CHILD(actualNodeID, vecinoNodeID);
								NODE_ID(newPointID) = actualNodeID;
								}
							else if (NODE(actualNodeID).inited && !NODE(vecinoNodeID).inited)
								// Directamente añadimos el otro nodo 
								{
								ADD_CHILD(vecinoNodeID, actualNodeID);
								NODE_ID(newPointID) = vecinoNodeID;
								}
							else if (NODE(actualNodeID).inited && NODE(vecinoNodeID).inited)
								// Tenemos dos nodos antiguos, y creamos un nodo padre para
								// unificarlos
								{
								uInt newNodeID;
								NEW_NODE(newNodeID, col, row, threshold);

								ADD_CHILD(newNodeID, actualNodeID);
								ADD_CHILD(newNodeID, vecinoNodeID);
								NODE_ID(newPointID) = newNodeID;
								}
							else // caso !NODE(actualNodeID).inited y !NODE(vecinoNodeID).inited
								// Tenemos dos nodos padre. Dejamos las cosas como están.
								// ¡No!, hay que unificar ambos, pasando los hijos de uno
								// al otro!
								{
								Q_ASSERT(NODE(actualNodeID).inited == false);
								Q_ASSERT(NODE(vecinoNodeID).inited == false);
								Q_ASSERT(NODE(actualNodeID).numChilds > 0);
								Q_ASSERT(NODE(vecinoNodeID).numChilds > 0);

								uInt child;
								for (	child = NODE(actualNodeID).child;
									NODE(child).brother != NULL_NODE;
									child = NODE(child).brother);

								uInt lastActualChildNodeID = NODE(actualNodeID).child;
								while (NODE(lastActualChildNodeID).brother != NULL_NODE)
									lastActualChildNodeID =
										NODE(lastActualChildNodeID).brother;

								NODE(actualNodeID).numChilds += NODE(vecinoNodeID).numChilds;
								NODE(lastActualChildNodeID).brother = NODE(vecinoNodeID).child;
								NODE_ID(newPointID) = actualNodeID;
								}
							}
						Q_ASSERT(NODE_ID(disjointSet.find(disjointSet.index(col, row))) == NODE_ID(disjointSet.find(disjointSet.index(i, j))));
						}
					}
			}

		// En este bucle actualizamos las areas para el corte de nivel threshold de los nodos antiguos, y creamos nodos
		// nuevos, caso de encontrar un conjunto de uno o varios píxeles de valor de gris igual al valor threshold,
		// que no estén unidos a ninguna región representada ya por un nodo en la estructura del arbol de componentes.
		//
		// En este punto, hemos procesado todos los píxeles de una misma intensidad de color, o menor.
		// Todos los píxeles están agrupados, bien con píxeles de un color igual, en cuyo caso tenemos el vértice de
		// una región, o bien anexos a una región cuyo vértice tiene una intensidad de color menor.
		//
		// Recorremos los puntos de intensidad igual, para bien actualizar áreas de las regiones a las que se han añadido,
		// para el threshold dado, o bien crear un nuevo nodo en el arbol de componentes y agruparlas en él.
		for (int n=0; n< points[threshold].size(); n++)
			{
			Q_ASSERT_X(threshold < 256, "getComponentTree", "out of bounds 4");
			QPoint p = points[threshold][n];
			int col = p.x(), row = p.y();
			uInt	actualID = disjointSet.index(col, row);

			Q_ASSERT_X(actualID < cols * rows, "getComponentTree", "out of bounds 5");

			// Tenemos un pixel que si bien tiene un valor de gris igual al threshold, su identificador dentro del
			// disjoint set es el mismo que el del grupo en el que está. Esto significa, o bien que es un píxel
			// aislado, rodeado de píxeles con valores de gris mayor a él, o bien que está en un grupo conexo de
			// píxeles con un valor de gris igual al suyo, rodeado a su vez por un conjunto de píxeles cuyo valor de
			// gris es mayor al suyo, y que ha sido elegido por el disjoint set como 'cabecera de grupo'.
			//
			// Crearemos un nodo a partir de él, cuya semilla será sus coordenadas, y no tendrá hijos, a priori.
			//
			if (actualID == disjointSet.find(actualID))
				{
				if (NODE_ID(actualID) == NULL_NODE)
					// Tenemos un punto que será cabecera en un nuevo nodo del arbol de componentes.
					// Inicializamos los valores de su nodo.
					{
					uInt newNode;
					NEW_NODE(newNode, col, row, threshold);
					this->valid[actualID] = true;
					this->leafNodes++;
					//qDebug() << "getComponentTree(): new node" << newNode << "Seed (" << col << row << ")";
					}
				else	// El pixel actual está ya asociado a un nodo, pero este nodo se ha creado en otro píxel.
					// Contamos como píxel libre aquel en que se creó el nodo que contiene el pixel actual.
					this->freePoints++;
				}
			else	// El pixel actual no es cabecera de su grupo.
				{
				this->freePoints++;
				// ¿Porque no?
				//Q_ASSERT(this->nodeID[actualID] == NULL_NODE);
				}
			uInt actualNodeID = NODE_ID(disjointSet.find(actualID));

			if (actualNodeID != NULL_NODE)
				{
				Q_ASSERT(NODE(actualNodeID).area[NODE(actualNodeID).lastThreshold]
						<= disjointSet.getSetCardinality(actualID));
				NODE(actualNodeID).lastThreshold = threshold;
				NODE(actualNodeID).area[threshold] = disjointSet.getSetCardinality(actualID);
				NODE(actualNodeID).inited = true;
				}

			// Actualizamos número total de puntos
			this->totalPoints++;
			}
		}

	this->rootNode = NODE_ID(disjointSet.find(0));

	// COMPONENT TREE FINISHED. STARTING TESTS...

	// At this point we have component tree constructed.
	Q_ASSERT(NODE(this->rootNode).inited);

	// TEST /////////////////////////////////////////////
	#ifndef QT_NO_DEBUG

	// Checking that actually number of nodes, and number of nodes with ID != NULL_NODE equals the
	// value holded in this->numNodes.
	uInt sumNodes = 0, sumChildNodes = 0, sumLeafNodes = 0;
	// Checking some basic tree structure and values
	for (uInt node = 0; node < this->numNodes; node++)
		if (NODE(node).inited)
			{
			uInt	firstThreshold = NODE(node).firstThreshold,
				lastThreshold = NODE(node).lastThreshold;

			if (NODE(node).numChilds == 0)
				sumLeafNodes++;

			sumNodes++;
			// Comprobamos que las áreas son correctas
			Q_ASSERT(NODE(node).area[firstThreshold] <= NODE(node).area[lastThreshold] != 0);
			Q_ASSERT(NODE(node).area[firstThreshold] != 0);
			Q_ASSERT(NODE(node).area[lastThreshold] != 0);

			// Comprobamos que el píxel de la imagen situado en la semilla, se
			// corresponde con el primer threshold
			Q_ASSERT(image(NODE(node).seedX, NODE(node).seedY) == firstThreshold);

			// Comprobamos que las áreas van de menor tamaño a mayor tamaño.
			uInt lastArea = NODE(node).area[firstThreshold];
			for (uInt threshold = firstThreshold+1; threshold <= lastThreshold; threshold++)
				if (NODE(node).area[threshold] != 0)
					{
					uInt actualArea = NODE(node).area[threshold];
					Q_ASSERT(actualArea >= lastArea);
					lastArea = actualArea;
					}

			uInt maxAreasChilds = 0;
			uInt numChilds = 0;
			for (	uint child = NODE(node).child;
				child != NULL_NODE;
				child = NODE(child).brother, numChilds++)
				{
				Q_ASSERT(numChilds < NODE(node).numChilds);
				Q_ASSERT(child < this->numNodes);
				Q_ASSERT(child != this->rootNode);
				Q_ASSERT(NODE(child).inited);
				maxAreasChilds += NODE(child).area[NODE(child).lastThreshold];
				}

			Q_ASSERT(NODE(node).area[NODE(node).firstThreshold] > maxAreasChilds);
			Q_ASSERT(numChilds == NODE(node).numChilds);
			// TODO: comprobar que la suma total de hijos más uno es exactamente el número de nodos.
			sumChildNodes += NODE(node).numChilds;
		}
	qDebug() << "getComponentTree(): /////// estadisticos";
	qDebug() << "getComponentTree(): sumNodes = " << sumNodes << sumChildNodes;
	Q_ASSERT( sumNodes == (sumChildNodes+1) );
	Q_ASSERT( this->leafNodes == sumLeafNodes );

	qDebug() << "getComponentTree(): this->numNodes = " << this->numNodes;
	qDebug() << "getComponentTree(): this->freePoints = "<< this->freePoints;
	qDebug() << "getComponentTree(): this->totalPoints = "<< this->totalPoints;
	qDebug() << "getComponentTree(): this->leafNodes + this->freePoints = "
			<< this->leafNodes + this->freePoints;

	Q_ASSERT(disjointSet.numberOfSets() == 1);
	Q_ASSERT(disjointSet.getSetCardinality(this->rootNode) == (image.getCols() * image.getRows()));
	Q_ASSERT(this->totalPoints == (image.getCols() * image.getRows()));
	Q_ASSERT(this->totalPoints == this->leafNodes + this->freePoints);

	#endif
	// TEST /////////////////////////////////////////////

	qDebug() << "getComponentTree() <~ return";
	}
}
